@@ -1,20 +1,24 @@
 import abc
 import warnings
 import pandas as pd
+import numpy as np
 
 from sklearn.base import is_classifier
 from sklearn.model_selection import train_test_split
 from sklearn.metrics import roc_auc_score
 from catboost import CatBoostClassifier
 from scipy import stats
-from collections import defaultdict
+from collections import defaultdict, namedtuple
 from typing import List
+from functools import partial
 
 from .interpretable_drift import InterpretableDrift
 from ..constants import RANDOM_STATE
 from ..exceptions import ColumnsNotMatchException
 from ..models import ScikitModel, cat_features_fillna
-from ..data import compute_levels_count_and_pct
+
+
+Coordinates = namedtuple('Coordinates', 'x0 x1 y0 y1')
 
 
 class DriftChecker(abc.ABC):
@@ -63,9 +67,9 @@ class DriftChecker(abc.ABC):
             raise ColumnsNotMatchException(
                 'Different columns for left and right dataframes\n\n'
                 f'Columns in right dataframe but not in left one: '
-                f'{(", ").join(column_name_right_not_in_left) or "None"}\n'
+                f'{", ".join(column_name_right_not_in_left) or "None"}\n'
                 f'Columns in left dataframe but not in right one: '
-                f'{(", ").join(column_name_left_not_in_right) or "None"}'
+                f'{", ".join(column_name_left_not_in_right) or "None"}'
             )
 
         self.df_left_data = df_left_data
@@ -88,7 +92,7 @@ class DriftChecker(abc.ABC):
     def ml_model_can_discriminate(self,
                                   ml_discriminate_model: ScikitModel = None,
                                   column_names: List[str] = None,
-                                  auc_threshold: float = .05,
+                                  auc_threshold: float = .1,
                                   new_target_column: str = 'is_left') -> bool:
         """Creates a machine learning model based in `sklearn`,
         this model will be a classification model that will try
@@ -150,7 +154,7 @@ class DriftChecker(abc.ABC):
             )
 
         X_train, X_test, y_train, y_test = train_test_split(
-            X, y, test_size=.2, random_state=RANDOM_STATE
+            X, y, test_size=.5, random_state=RANDOM_STATE
         )
 
         self.ml_discriminate_model.fit(X_train, y_train)
@@ -162,15 +166,17 @@ class DriftChecker(abc.ABC):
         self.drift = (symmetric_auc(auc_drift_check_model)
                       < symmetric_auc(auc_threshold))
 
-        if not self.minimal:
-            self.interpretable_drift = InterpretableDrift(
-                model=self.ml_discriminate_model,
-                X_train=X_train,
-                X_test=X_test,
-                column_names=(column_names if column_names
-                              else self.df_left_data.columns.tolist())
-            )
+        self.interpretable_drift = InterpretableDrift(
+            model=self.ml_discriminate_model,
+            X_train=X_train,
+            X_test=X_test,
+            y_train=y_train,
+            y_test=y_test,
+            column_names=(column_names if column_names
+                          else self.df_left_data.columns.tolist())
+        )
 
+        if not self.minimal:
             self.interpretable_drift.most_discriminative_features_plot()
 
         is_there_drift = symmetric_auc(auc_drift_check_model) > auc_threshold
@@ -199,18 +205,16 @@ class DataDriftChecker(DriftChecker):
                  df_right_data: pd.DataFrame,
                  verbose: bool = True,
                  minimal: bool = False,
-                 pvalue_threshold: float = .01,
-                 cardinality_threshold: int = 20,
-                 pct_level_threshold: float = .05,
-                 pct_change_level_threshold: float = .05):
+                 pvalue_threshold_numerical: float = .05,
+                 pvalue_threshold_categorical: float = .05,
+                 cardinality_threshold: int = 20):
         """Inits `DriftChecker` with `left_data` and `right_data`
         to compare
 
         Both needed in `pandas.DataFrame` type
 
-        `pvalue_threshold` is set to .01 by default, `cardinality_threshold`
-        is set to 20 , `pct_level_threshold` is set to .05 and
-        `pct_change_level_threshold` is set to .05, you can modify
+        `pvalue_threshold` is set to .05 by default, `cardinality_threshold`
+        is set to 20, and `pct_level_threshold` is set to .05, you can modify
         them if your problem fits better with other thresholds
 
         This class refers to data drift checks, which what it compares
@@ -218,27 +222,139 @@ class DataDriftChecker(DriftChecker):
         """
         super().__init__(df_left_data, df_right_data, verbose, minimal)
 
-        self.pvalue_threshold = pvalue_threshold
+        self.pvalue_threshold_numerical = pvalue_threshold_numerical
+        self.pvalue_threshold_categorical = pvalue_threshold_categorical
         self.cardinality_threshold = cardinality_threshold
-        self.pct_level_threshold = pct_level_threshold
-        self.pct_change_level_threshold = pct_change_level_threshold
+        self.dict_each_column_pvalues_numerical = defaultdict(float)
+        self.dict_each_column_pvalues_categorical = defaultdict(float)
+        self.dict_each_column_cardinality = defaultdict(float)
+        self.dict_each_column_drift_coefficients = defaultdict(float)
+        self.high_cardinality_features = []
+        self.drifted_features = set()
+        self._convert_pvalues_to_drift_coefficients_numerical = partial(
+            self._convert_pvalues_to_drift_coefficients,
+            pvalue_threshold=self.pvalue_threshold_numerical
+        )
+        self._convert_pvalues_to_drift_coefficients_categorical = partial(
+            self._convert_pvalues_to_drift_coefficients,
+            pvalue_threshold=self.pvalue_threshold_categorical
+        )
+
+    @staticmethod
+    def _line_equation(x: float, points_coordinates: Coordinates) -> float:
+        """Computes the line equation given two points
+        `points_coordinates` and returns `y` value for
+        specific `x`
+        """
+        def slope_from_coordinates(
+                _points_coordinates: Coordinates) -> float:
+            """Slope equation:
+
+                (y0 - y1) / (x0 - x1)
+            """
+            return ((_points_coordinates.y0 - _points_coordinates.y1)
+                    / (_points_coordinates.x0 - _points_coordinates.x1))
+
+        def intercept_from_coordinates(
+                _points_coordinates: Coordinates) -> float:
+            """Intercept equation:
+
+                (x0 * y1 - x1 * y0) / (x0 - x1)
+            """
+            return ((_points_coordinates.x0 * _points_coordinates.y1
+                     - _points_coordinates.x1 * _points_coordinates.y0)
+                    / (_points_coordinates.x0 - _points_coordinates.x1))
+
+        return (slope_from_coordinates(points_coordinates) * x
+                + intercept_from_coordinates(points_coordinates))
+
+    def _convert_pvalues_to_drift_coefficients(
+            self, pvalue: float, pvalue_threshold: float) -> float:
+        """
+        self.
+        """
+
+        dict_coordinates = (
+            dict(x0=0,
+                 x1=pvalue_threshold,
+                 y0=0,
+                 y1=0.5) if pvalue < pvalue_threshold
+            else dict(x0=pvalue_threshold,
+                      x1=1,
+                      y0=.5,
+                      y1=1)
+        )
+
+        coordinates = Coordinates(**dict_coordinates)
+
+        return 1 - self._line_equation(pvalue, coordinates)
+
+    def _update_each_column_drift_coefficient(self) -> None:
+        """Updates `self.dict_each_column_pvalues` when
+        all of the pvalues are calculated
+        """
+        are_all_features_pvalues_filled = (
+                self.dict_each_column_pvalues_numerical
+                and self.dict_each_column_pvalues_categorical
+        )
+
+        if are_all_features_pvalues_filled:
+            dict_each_column_drift_coefficients_numerical = {
+                feature: (
+                    self
+                    ._convert_pvalues_to_drift_coefficients_numerical(
+                        pvalue
+                    )
+                )
+                for feature, pvalue in (self
+                                        .dict_each_column_pvalues_numerical
+                                        .items())
+            }
+
+            dict_each_column_drift_coefficients_categorical = {
+                feature: (
+                    self
+                    ._convert_pvalues_to_drift_coefficients_categorical(
+                        pvalue
+                    )
+                )
+                for feature, pvalue in (self
+                                        .dict_each_column_pvalues_categorical
+                                        .items())
+            }
+
+            self.dict_each_column_drift_coefficients = {
+                **dict_each_column_drift_coefficients_numerical,
+                **dict_each_column_drift_coefficients_categorical
+            }
 
     def check_numerical_columns(self) -> bool:
         """Given `numerical_columns` check all drifts
+
+        Kolmogorov-Smirnov test:
+
+            This is a two-sided test for the null hypothesis that 2
+            independent samples are drawn from the same continuous
+            distribution.
         """
-        dict_each_column_pvalues = defaultdict(float)
         for numerical_column in self.num_features:
             _, pvalue = stats.ks_2samp(self.df_left_data[numerical_column],
                                        self.df_right_data[numerical_column])
-            dict_each_column_pvalues[numerical_column] = pvalue
+            self.dict_each_column_pvalues_numerical[numerical_column] = pvalue
+
+        self._update_each_column_drift_coefficient()
 
         drifted_features = [
             column
-            for column, pvalue in dict_each_column_pvalues.items()
-            if pvalue < self.pvalue_threshold
+            for column, pvalue in (self
+                                   .dict_each_column_pvalues_numerical
+                                   .items())
+            if pvalue < self.pvalue_threshold_numerical
         ]
 
-        is_there_drift = len(drifted_features) > 0
+        self.drifted_features = self.drifted_features.union(drifted_features)
+
+        is_there_drift = len(self.drifted_features) > 0
 
         if self.verbose:
             if is_there_drift:
@@ -248,7 +364,8 @@ class DataDriftChecker(DriftChecker):
                       'otherwise check the data source',
                       end='\n\n')
 
-                print(f'Features drifted: {(", ").join(drifted_features)}')
+                warnings.warn(f'Features drifted (numerical): '
+                              f'{", ".join(drifted_features)}')
             else:
                 print('No drift found in numerical columns check step',
                       end='\n\n')
@@ -257,53 +374,77 @@ class DataDriftChecker(DriftChecker):
 
     def check_categorical_columns(self) -> bool:
         """Given `categorical_columns` check all drifts
+
+        Calculate a one-way chi-square test:
+
+            The chi-square test tests the null hypothesis
+            that the categorical data has the given frequencies.
         """
-        dict_each_column_cardinality = defaultdict(float)
-        dict_each_column_level_difference = defaultdict(float)
+        sample_size = min(len(self.df_left_data), len(self.df_right_data))
+
         for categorical_column in self.cat_features:
-            categorical_levels_left = compute_levels_count_and_pct(
-                self.df_left_data, categorical_column=categorical_column
+            df_left_data_to_compare = (self
+                                       .df_left_data[categorical_column]
+                                       .sample(sample_size))
+
+            df_right_data_to_compare = (self
+                                        .df_right_data[categorical_column]
+                                        .sample(sample_size))
+
+            # .1 is just to avoid zero-division in some cases
+            dict_left_frequencies = (
+                df_left_data_to_compare
+                .value_counts()
+                .to_dict(defaultdict(lambda: .01))
             )
 
-            categorical_levels_right = compute_levels_count_and_pct(
-                self.df_right_data, categorical_column=categorical_column
+            dict_right_frequencies = (
+                df_right_data_to_compare
+                .value_counts()
+                .to_dict(defaultdict(lambda: .01))
             )
 
-            categorical_levels_joined = (
-                categorical_levels_left
-                .merge(categorical_levels_right,
-                       on='index',
-                       suffixes=('_left', '_right'),
-                       how='outer')
+            frequencies_zipped = [
+                (dict_left_frequencies[key], dict_right_frequencies[key])
+                for key in (set(dict_left_frequencies)
+                            .union(set(dict_right_frequencies)))
+            ]
+
+            _, pvalue = stats.chisquare(
+                list(map(lambda x: x[0], frequencies_zipped)),
+                f_exp=list(map(lambda x: x[1], frequencies_zipped))
             )
 
-            dict_each_column_cardinality[categorical_column] = (
-                len(categorical_levels_joined)
+            self.dict_each_column_cardinality[categorical_column] = (
+                len(frequencies_zipped)
             )
 
-            pct_category_level_left = f'{categorical_column}_norm_left'
-            pct_category_level_right = f'{categorical_column}_norm_right'
+            self.dict_each_column_pvalues_categorical[
+                categorical_column
+            ] = pvalue
 
-            dict_each_column_level_difference[categorical_column] = (
-                abs(categorical_levels_joined[pct_category_level_left]
-                    - categorical_levels_joined[pct_category_level_right])
-                .iloc[0]
-            )
+        self._update_each_column_drift_coefficient()
 
-        high_cardinality_features = [
+        self.high_cardinality_features = [
             column
-            for column, cardinality in dict_each_column_cardinality.items()
+            for column, cardinality in (self
+                                        .dict_each_column_cardinality
+                                        .items())
             if cardinality > self.cardinality_threshold
         ]
 
         drifted_features = [
             column
-            for column, difference in dict_each_column_level_difference.items()
-            if difference > self.pct_change_level_threshold
+            for column, pvalue in (self
+                                   .dict_each_column_pvalues_categorical
+                                   .items())
+            if pvalue < self.pvalue_threshold_categorical
         ]
 
+        self.drifted_features = self.drifted_features.union(drifted_features)
+
         is_there_drift = len(drifted_features) > 0
-        is_there_high_cardinality = len(high_cardinality_features)
+        is_there_high_cardinality = len(self.high_cardinality_features)
 
         if self.verbose:
             if is_there_drift:
@@ -313,15 +454,15 @@ class DataDriftChecker(DriftChecker):
                       'otherwise check the data source',
                       end='\n\n')
 
-                warnings.warn(f'Features drifted: '
-                              f'{(", ").join(drifted_features)}')
+                warnings.warn(f'Features drifted (categorical): '
+                              f'{", ".join(drifted_features)}')
             else:
                 print('No drift found in categorical columns check step',
                       end='\n\n')
 
             if is_there_high_cardinality:
                 warnings.warn(f'Features cardinality warning: '
-                              f'{(", ").join(high_cardinality_features)}')
+                              f'{", ".join(self.high_cardinality_features)}')
 
         return is_there_drift
 
@@ -358,8 +499,11 @@ class ModelDriftChecker(DriftChecker):
         self.target_column_name = target_column_name
         self.auc_threshold = auc_threshold
         self.interpretable_drift_classifier_model = None
+        self.y_score_left = None
 
-    def check_model(self, column_names: List[str] = None) -> bool:
+    def check_model(self,
+                    column_names: List[str] = None,
+                    new_target_column: str = 'is_left') -> bool:
         """Checks if features relations with target are the same
         for `self.df_left_data` and `self.df_right_data`
 
@@ -372,10 +516,15 @@ class ModelDriftChecker(DriftChecker):
         X_right = self.df_right_data.drop(columns=self.target_column_name)
         y_right = self.df_right_data[self.target_column_name]
 
-        y_score_left = self.ml_classifier_model.predict_proba(X_left)[:, 1]
-        y_score_right = self.ml_classifier_model.predict_proba(X_right)[:, 1]
+        self.y_score_left = (self
+                             .ml_classifier_model
+                             .predict_proba(X_left)[:, 1])
 
-        auc_left = roc_auc_score(y_true=y_left, y_score=y_score_left)
+        y_score_right = (self
+                         .ml_classifier_model
+                         .predict_proba(X_right)[:, 1])
+
+        auc_left = roc_auc_score(y_true=y_left, y_score=self.y_score_left)
         auc_right = roc_auc_score(y_true=y_right, y_score=y_score_right)
 
         if not self.minimal:
@@ -383,8 +532,20 @@ class ModelDriftChecker(DriftChecker):
                 model=self.ml_classifier_model,
                 X_train=X_left,
                 X_test=X_right,
+                y_train=(
+                    pd
+                    .Series(1, name=new_target_column)
+                    .repeat(len(X_left))
+                    .to_frame()
+                ),
+                y_test=(
+                    pd
+                    .Series(0, name=new_target_column)
+                    .repeat(len(X_right))
+                    .to_frame()
+                ),
                 column_names=(column_names if column_names
-                              else self.df_left_data.columns.tolist())
+                              else (X_left.columns.tolist()))
             )
 
             (self
@@ -405,7 +566,76 @@ class ModelDriftChecker(DriftChecker):
                 end='\n\n'
             )
 
-            print(f'AUC left data: {auc_left}')
-            print(f'AUC right data: {auc_right}')
+            print(f'AUC left data: {auc_left:.2f}')
+            print(f'AUC right data: {auc_right:.2f}')
 
         return is_there_drift
+
+    def show_feature_importance_vs_drift_map_plot(self, top: int = 10) -> None:
+        """Shows feature importance versus drift coefficient
+        map
+
+        By default shows you the top 10 most important features
+        but you can customize it with `top` parameter
+
+        You can found more details for this function in its module:
+
+            `pydrift.InterpretableDrift.feature_importance_vs_drift_map_plot`
+        """
+        data_drift_checker = DataDriftChecker(self.df_left_data,
+                                              self.df_right_data,
+                                              verbose=False,
+                                              minimal=True)
+
+        data_drift_checker.check_numerical_columns()
+        data_drift_checker.check_categorical_columns()
+
+        (self
+         .interpretable_drift_classifier_model
+         .feature_importance_vs_drift_map_plot(
+            dict_each_column_drift_coefficient=(
+                data_drift_checker.dict_each_column_drift_coefficients
+            ),
+            top=top))
+
+    def sample_weight_for_retrain(self) -> np.array:
+        """If you need to retrain your model maybe
+        it's better applying this weights when you
+        do it
+
+        From https://bit.ly/2Xf39ks (thanks!):
+
+            We can use this w as sample weights in any of our
+            classifier to increase the weight of these observation
+            which seems similar to our test data. Intuitively this
+            makes sense as our model will focus more on capturing
+            patterns from the observations which seems similar to our test.
+
+        An example with random forest:
+
+            rf = RandomForestClassifier(**rf_params)
+            rf.fit(X_train, y_train, sample_weight=weights)
+        """
+        # Temporary change `self.minimal` to avoid confusing plots
+        actual_self_minimal, self.minimal = self.minimal, True
+        # Temporary change `self.verbose` to avoid confusing plots
+        actual_verbose, self.verbose = self.verbose, False
+
+        self.ml_model_can_discriminate()
+
+        # Reset verbose and minimal values
+        self.minimal = actual_self_minimal
+        self.verbose = actual_verbose
+
+        weights = (1 / self.y_score_left) - 1
+        weights /= np.mean(weights)
+
+        if not self.minimal:
+            (self
+             .interpretable_drift
+             .weights_plot(weights))
+
+            print('Higher the weight for the observation, '
+                  'more is it similar to the test data')
+
+        return weights
